/***************************************************************************
 *                          Zanoza Modeler v2.x                            *
 *      This is unpublished propietary source code of Zanoza Software.     *
 *      The copyright notice does not evidence any actual or intended      *
 *                    publication of such source code.                     *
 *                                                                         *
 *                Copyright (c) 2002-2003 Zanoza Software.                 *
 *                                                                         *
 *                        All rights reserved.                             *
 ***************************************************************************
 *--------------------------------------------------------------------------
 * @author:   Oleg M.
 *--------------------------------------------------------------------------
 * @purpose:  Spline class implementation
 *--------------------------------------------------------------------------
 * @histroy: 16.09.2003. Created
 */
#include "Spline.h"

namespace scene
{

INTERFACE_BEGINMAP(CSpline)
  INTERFACE_ENTRY(scene::ISpline)
  INTERFACE_ENTRY(core::io::ISerializeable)
  INTERFACE_ENTRY(core::IErrorInfo)
INTERFACE_ENDMAP()

//global:
float g_fPrecision = 0.95f; //[0.0f..1.0f] with 1.0f as maximum

CSpline::CSpline() : CErrorInfo()
{
  m_bClosed = false;
}


CSpline::~CSpline()
{
}

ZRESULT CSpline::isClosed(void)
{
  return m_bClosed ? ZRESULT_OK : ZRESULT_FALSE;
}


//-------------------------------------------------------
// @name : CSpline::setClosed
// Sets closed flag
// @param bool bClosed                   : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSpline::setClosed(bool bClosed)
{
  m_bClosed = bClosed;
  return ZRESULT_OK;
}



//-------------------------------------------------------
// @name : CSpline::getSectionControls
// Provides section end-control point in respect to closed splines. ZRESULT_FALSE is
// returned once last section achieved.
// @param scene::IControlPoint* pStart   : 
// @param scene::IControlPoint** ppEnd   : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSpline::getSectionControls(scene::IControlPoint* pStart, scene::IControlPoint** ppEnd)
{
  resetError();
  if (NULL == pStart || NULL == ppEnd)
    return ZRESULT_INVALID_ARG;
  pushState();//navigation can distort current positions
  if (ZRESULT_OK != navigateTo(pStart))
    return REPORT_ERROR(NULL, _T("Supplied control point was not found. "))
  ZPtr<IControlPoint> p;
  next(NULL);//pStart
  next(&p);  //one after pStart
  bool bEnd = !current();
  bool bDone = false;
  if (!p && m_bClosed)
  {
    reset();
    next(&p);
    bDone = true;
  }
  p.CopyTo(ppEnd);
  popState();
  return (bDone || !m_bClosed && bEnd) ? ZRESULT_FALSE : ZRESULT_OK;
}


//-------------------------------------------------------
// @name : CSpline::CSpline::getCurrentSectionPoint
// Provides point for current section using supplied u [0..1]
// @param float u                        : 
// @param ZPoint3D& pt                   : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSpline::getCurrentSectionPoint(float u, ZPoint3D& pt)
{
  resetError();
  if (!current())
    return REPORT_ERROR(NULL, _T("Empty or non-reset spline iterator. "));
  zmath::clamp(u, 0.0f, 1.0f);
  ZPtr<IControlPoint> pStart, pEnd;
  next(&pStart);
  if (ZRESULT_OK != next(&pEnd))
  {
    reset();
    next(&pEnd);
  }
  return getSectionPoint(pStart, pEnd, u, pt);
}


//-------------------------------------------------------
// @name : CSpline::getSectionLength
// Provides length of supplied section
// @param scene::IControlPoint* pS       : 
// @param scene::IControlPoint* pE       : 
// @param float& fLen                    : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSpline::getSectionLength(scene::IControlPoint* pS, scene::IControlPoint* pE, float& fLen)
{
  fLen = 0.0f;
  if (NULL == pS || NULL == pE)
    return ZRESULT_FALSE;
  ZPoint3D ptS, ptE;
  ZVector  vS,  vE;
  bool bAuto = false;
  long nDet = 0;
  pE->getControlPosition(ptE);
  pE->getControlVector(scene::SCV_U1, vE);
  pS->getControlPosition(ptS);
  pS->getControlVector(scene::SCV_U0, vS);
  pS->getDetalization(nDet, &bAuto);
  ZPoint3D ptLen = ptS;
  ZPoint3D ptLen2;
  for (int i = 1; i <= nDet+1; i++)
  {
    ptLen2 = getSectionPoint(ptS, ptE, vS, vE, (float)i/(float)(nDet+1));
    fLen += ptLen.distance(ptLen2);
    ptLen = ptLen2;
  }
  return ZRESULT_OK;
}

//-------------------------------------------------------
// @name : CSpline::getSplinePoint
// Provides point for the spline using supplied u [0..1]
// @param float u                        : 
// @param ZPoint3D& pt                   : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSpline::getSplinePoint(float u, float fLength, ZPoint3D& pt)
{
  resetError();
  zmath::clamp(u, 0.0f, 1.0f);
  if (fLength <= 0)
  {
    getSplineLength(fLength);
    if (fLength <= 0)
      return REPORT_ERROR(NULL, _T("Invalid detail param for CSpline::getSplinePoint."));
  }
  float fLim = fLength*u;
  float fLen = 0.0f;
  float fSec = 0.0f;
  ZPtr<IControlPoint> p;
  ZPtr<IControlPoint> pLast;
  reset();
  while (ZRESULT_OK == next(&p))
  {
    if (pLast)
      getSectionLength(pLast, p, fSec);
    if (fLen + fSec >= fLim)
      break;
    else
      fLen += fSec;
    fSec = 0.0f;
    pLast = p;
    p = NULL;
  }
  if (!p && pLast && m_bClosed && (fLen + fSec <= fLim))
  {
    //
    // last point autodetail
    //
    reset();
    next(&p);
    getSectionLength(pLast, p, fSec);
  }
  if (pLast)
  {
    if ((!p && pLast) && 
        ((fabs(fLen+fSec-fLim) < MINFLOAT) || (fabs(fSec) < MINFLOAT)))
      return pLast->getControlPosition(pt);
    else
    {
      float u2 = (fLim-fLen)/fSec;
      return getSectionPoint(pLast, p, u2, pt);
    }
    return ZRESULT_OK;
  }
  else
    if (p)
      return p->getControlPosition(pt);

  return ZRESULT_FALSE;
}

//-------------------------------------------------------
// @name : CSpline::getSplineControlPoint
// Provides a control point which controls the curve pointed by u-parameter [0..1]
// @param float u                        : 
// @param scene::IControlPoint** ppPoint : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSpline::getSplineControlPoint(float u, scene::IControlPoint** ppPoint)
{
  resetError();
  zmath::clamp(u, 0.0f, 1.0f);
  long  nCurvePt = numCurvePoints(false);
  float f = 1.0f/(float)nCurvePt;
  reset();
  ZPtr<IControlPoint> p;
  nCurvePt = 0;
  while (ZRESULT_OK == next(&p))
  {
    long n;
    p->getDetalization(n);
    nCurvePt += n+1;
    if (f*nCurvePt >= u)
      break;
    p = NULL;
  }
  if (!p)
    return REPORT_ERROR1(NULL, _T("Can't obtain control point for supplied u: %.4f. "), u);
  return p.CopyTo(ppPoint);
}


//-------------------------------------------------------
// @name : CSpline::getAutoDetail
// Provides a detalization for supplied segment
// @param ZPoint3D& ptS                  : 
// @param ZPoint3D&ptE                   : 
// @param ZVector& vS                    : 
// @param ZVector& vE                    : 
//
// @return long : 
//-------------------------------------------------------
long CSpline::getAutoDetail(ZPoint3D& ptS, ZPoint3D&ptE, ZVector& vS, ZVector& vE)
{
  ZPoint3D ptCurveMid = getSectionPoint(ptS, ptE, vS, vE, 0.5f);
  ZLine3D  segment(ptS, ptE);
  float distleft = ptCurveMid.distance2(ptS);
  float distright= ptCurveMid.distance2(ptE);
  float distsegment = segment.distance(ptCurveMid);
  float sL = vS.length();
  float eL = vE.length();
  float L  = ptS.distance(ptE);
  float fDistPrecision = MINFLOAT*(1.0f-g_fPrecision)*100000.0f;//[MINFLOAT..100k*MINFLOAT]
  float fAngPrecision = MINFLOAT*(1.0f-g_fPrecision)*10000.0f;//[MINFLOAT..10k*MINFLOAT]
  bool  bSegmentAffect = (sL > fDistPrecision && ((vS^segment.direction())/(sL*L) > 1.0f-fAngPrecision)) ||
                         (eL > fDistPrecision && ((vE^segment.direction())/(eL*L) > 1.0f-fAngPrecision));
  if (distleft < fDistPrecision ||
      distright < fDistPrecision ||
      (sL < MINFLOAT && eL < MINFLOAT) ||
      distsegment < fDistPrecision && bSegmentAffect)
    return 0;
  //otherwise: detalization requires
  ZVector vS1 = vS*0.5f;
  ZVector vE1 = (ptS+vS-ptCurveMid)*0.5f;
  ZVector vS2 = (ptE+vE-ptCurveMid)*0.5f;
  ZVector vE2 = vE*0.5f;
  return  min(g_nMaxDetail, 1 + 
                          getAutoDetail(ptS, ptCurveMid, vS1, vE1) + //left  segment
                          getAutoDetail(ptCurveMid, ptE, vS2, vE2));  //right segment
}


//-------------------------------------------------------
// @name : CSpline::numCurvePoints
// when bControl true: provides amount of control points in spline; otherwise: provides
// complete amount of points in the spline (using section detalization values).
// @param bool bControl                  : 
//
// @return long : 
//-------------------------------------------------------
long CSpline::numCurvePoints(bool bControl)
{
  resetError();
  if (bControl)
    return length();
  bool bAuto = false;
  long num = 0;
  long det = 0;
  reset();
  ZPtr<IControlPoint> p;
  ZPtr<IControlPoint> pLast;
  ZPoint3D ptLast, ptThis;
  ZVector  vLast,  vThis;
  while (ZRESULT_OK == next(&p))
  {
    p->getControlPosition(ptThis);
    p->getControlVector(scene::SCV_U1, vThis);
    if (bAuto && det < 0)
    {
      //previouse point:
      det = getAutoDetail(ptLast, ptThis, vLast, vThis);
      pLast->setDetalization(det, true);
    }
    num+= 1+det;//previouse point and previouse detalization
    p->getDetalization(det, &bAuto);
    ptLast = ptThis;
    p->getControlVector(scene::SCV_U0, vLast);
    pLast = p;
    p = NULL;
  }
  if (m_bClosed)
  {
    if (bAuto && det < 0)
    {
      //
      // last point autodetail
      //
      reset();
      next(&p);
      p->getControlPosition(ptThis);
      p->getControlVector(scene::SCV_U1, vThis);
      det = getAutoDetail(ptLast, ptThis, vLast, vThis);
      pLast->setDetalization(det, true);
    }
    num += 1+det;//last section which goes to the head.
  }
  return num;
}


//-------------------------------------------------------
// @name : CSpline::getSectionAutoDetail
// Provides adaptive sectino auto-detail;
// @param scene::IControlPoint* pStart   : 
// @param scene::IControlPoint* pEnd     : 
//
// @return long : 
//-------------------------------------------------------
long CSpline::getSectionAutoDetail(scene::IControlPoint* pStart, scene::IControlPoint* pEnd)
{
  if (NULL == pStart || NULL == pEnd)
    return 0;
  ZPoint3D ptS, ptE;
  ZVector  vS, vE;
  pStart->getControlPosition(ptS);
  pStart->getControlVector(scene::SCV_U0, vS);
  pEnd->getControlPosition(ptE);
  pEnd->getControlVector(scene::SCV_U1, vE);
  return getAutoDetail(ptS, ptE, vS, vE);
}


ZRESULT CSpline::getSectionPoint(scene::IControlPoint* pStart, scene::IControlPoint* pEnd, float u, ZPoint3D& pt)
{
  resetError();
  if (NULL == pStart || NULL == pEnd)
    return REPORT_ERROR(NULL, _T("Invalid Section supplied for CSpline::getSectionPoint."));
  ZPoint3D ptS, ptE;//start/ent
  pStart->getControlPosition(ptS);
  pEnd->getControlPosition(ptE);
  ZVector vCS, vCE;
  eControlPoint tS, tE;
  pStart->getType(tS);
  pEnd->getType(tE);
  if (tS != SCP_CORNER)
    pStart->getControlVector(SCV_U0, vCS);
  if (tE != SCP_CORNER)
    pEnd->getControlVector(SCV_U1, vCE);
  zmath::clamp(u, 0.0f, 1.0f);
  float u0 = u*u*u;
  float u1 = u*u*(1.0f-u);
  float u2 = u*(1.0f-u)*(1.0f-u);
  float u3 = (1.0f-u)*(1.0f-u)*(1.0f-u);

  pt = ptS*u3 + 3.0f*(ptS+vCS)*u2 + 3.0f*(ptE+vCE)*u1 + ptE*u0;
  return ZRESULT_OK;
}


//-------------------------------------------------------
// @name : CSpline::getSplineLength
// Provides length of spline
// @param float& fLen                     : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSpline::getSplineLength(float& fLen)
{
  resetError();
  fLen = 0.0f;
  float fSec = 0.0f;
  reset();
  ZPtr<IControlPoint> p;
  ZPtr<IControlPoint> pLast;
  while (ZRESULT_OK == next(&p))
  {
    if (ZRESULT_OK == getSectionLength(pLast, p, fSec))
      fLen+=fSec;
    pLast = p;
    p = NULL;
  }
  if (m_bClosed)
  {
    //
    // last point autodetail
    //
    reset();
    next(&p);
    if (ZRESULT_OK == getSectionLength(pLast, p, fSec))
      fLen+=fSec;
  }
  return ZRESULT_OK;
}

//helpers

//-------------------------------------------------------
// @name : CSpline::getSectionPoint
// Provides section point for supplied 
// @param IControlPoint* p               : 
// @param float u                        : 
//
// @return ZPoint3D : 
//-------------------------------------------------------
/*
ZPoint3D CSpline::getSectionPoint(IControlPoint* p, float u)
{
  resetError();
  ZPoint3D pt;
  if (NULL == p)
    return pt;
  ZPtr<IControlPoint> ptNext;
  //we are assuming we are already navigted
  //if (ZRESULT_OK != navigateTo(p))
  //  return pt; //missing control point?
  if (ZRESULT_OK != next(&ptNext))
  {
    reset();
    next(&ptNext);
  }
  if (!ptNext)
    return pt;
  ZPoint3D ptS, ptE;//start/ent
  p->getControlPosition(ptS);
  ptNext->getControlPosition(ptE);
  ZVector vCS, vCE;
  eControlPoint tS, tE;
  p->getType(tS);
  ptNext->getType(tE);
  if (tS != SCP_CORNER)
    p->getControlVector(SCV_U0, vCS);
  if (tE != SCP_CORNER)
    ptNext->getControlVector(SCV_U1, vCE);
  zmath::clamp(u, 0.0f, 1.0f);
  float u0 = u*u*u;
  float u1 = u*u*(1.0f-u);
  float u2 = u*(1.0f-u)*(1.0f-u);
  float u3 = (1.0f-u)*(1.0f-u)*(1.0f-u);

  return ptS*u3 + 3.0f*(ptS+vCS)*u2 + 3.0f*(ptE+vCE)*u1 + ptE*u0;
}
*/

};//scene

